# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
import os
import logging
import bzr
import phase
import shutil
import shlex
import re
import hashlib
import sys
import pwd
import errno
import json
from .contain import (
    LXCContainer,
    NoContainer,
    NoSuchContainerError,
)
from .utils import (
    needs_root,
    mkdirs,
    mksymlink,
    CacheDisk,
)

MOJO_ROOT = "/srv/mojo"

RESERVED_PROJECT_NAMES = [
    'LOCAL',
    'bin',
    'lib',
]


#######################################
# Spec
#######################################

class SpecUninitializedError(Exception):
    pass


class ConfigNotFoundException(Exception):
    def __init__(self, message, search_list):
        self.message = message
        self.search_list = search_list


class Spec(object):
    def __init__(self, spec_dir, remote=None, stage=None, manifest_file=None):
        self.spec_dir = spec_dir
        self.manifest_file = manifest_file
        self.local_branch = bzr.BzrBranch(self.spec_dir)
        if remote:
            self.remote_branch = bzr.BzrBranch(remote)
        elif self.is_initialized:
            if self.local_branch.parent:
                self.remote_branch = bzr.BzrBranch(self.local_branch.parent)
            else:
                self.remote_branch = None
        else:
            raise ValueError("No remote or local branch specified for Spec")
        self.stage = stage
        self._manifest = None

    @property
    def is_initialized(self):
        return self.local_branch is not None and self.local_branch.is_branch

    def initialize(self, clean_spec=True):
        "Download or update spec on disk"
        if not os.path.isdir(self.spec_dir):
            os.makedirs(self.spec_dir)
        if self.local_branch.is_branch:
            self.local_branch.pull(src_location=self.remote_branch.location,
                                   overwrite=clean_spec)
        else:
            self.local_branch = self.remote_branch.clone(self.spec_dir)
        if clean_spec:
            self.local_branch.revert()
            self.local_branch.cleantree()

    @property
    def manifest(self):
        logging.info("Retrieve the spec's manifest")
        if self._manifest:
            return self._manifest
        if not self.is_initialized:
            return None
        self._manifest = Manifest(self, manifest_file=self.manifest_file,
                                  stage=self.stage)
        return self._manifest

    def _remove_duplicates(self, mylist):
        """Remove duplicates while preserving order.

        See http://stackoverflow.com/a/480227/523729
        """
        seen = set()
        seen_add = seen.add
        return [
            x for x in mylist
            if not (x in seen or seen_add(x))]

    def iterate_subpaths(self, path):
        """Iterate subpaths of a path, from most to least explicit.
        e.g. input of 'a/b/c' will yield ('a/b/c', 'b/c', 'c')"""
        if path:
            yield path
            parts = path.split('/', 1)
            if len(parts) == 1:
                yield parts[0]
            else:
                for _path in self.iterate_subpaths(parts[1]):
                    yield _path

    def get_configs(self, base, stage=None):
        """Locate a list of stage-qualified configuration files.
           Configs are returned relative to the spec dir from
           least specific to most specfic in this order:
             * <base>
             * <stage>/../<base>
             * <stage>/<base>"""
        if not self.is_initialized:
            raise SpecUninitializedError("Spec has no local copy: {}"
                                         "".format(self.spec_dir))
        config_names = []
        if stage:
            # Common edge case $MOJO_STAGE/../$CONFIG
            config_names.append(os.path.join(stage, '..', base))
            for path in self.iterate_subpaths(os.path.join(stage, base)):
                config_names.append(path)
        else:
            for path in self.iterate_subpaths(base):
                config_names.append(path)

        config_names = [os.path.normpath(config) for config in config_names]
        config_names = self._remove_duplicates(config_names)
        # Order is important here
        # Least specific to most specific
        # Sort by string length
        config_names.sort(key=len)

        configs = [
            config for config in config_names
            if os.path.isfile(os.path.join(self.spec_dir, config))]

        if configs:
            return configs

        raise ConfigNotFoundException(
            "Config file '{}' not found.\n"
            "Is the MOJO_STAGE environment variable or the --stage option "
            "set?\nIs {} a spec branch?".format(base,
                                                self.local_branch.location),
            config_names)

    def get_config(self, base, stage=None):
        """Locate a stage-qualified configuration file.
           The absolute path to the most specific match,
           the first of the following configs is returned:
             * <stage>/<base>
             * <stage>/../<base>
             * <base>"""
        # The last element is the most specific matching path
        return os.path.join(self.spec_dir, self.get_configs(base, stage)[-1])


#######################################
# Manifest
#######################################

class InvalidManifestException(Exception):
    pass


class Manifest(object):
    def __init__(self, spec, manifest_file, stage):
        super(Manifest, self).__init__()
        self.spec = spec
        self.stage = stage
        self.config = spec.get_config(base=manifest_file, stage=stage)
        if manifest_file == "manifest":
            self.required_commands = ('collect', 'deploy')
        else:
            self.required_commands = ()
        self.phases = []
        self.load()

    def load(self):
        self.phases = self.parse_manifest(self.config)
        commands = [phase.name for phase in self.phases]
        for command in self.required_commands:
            if command not in commands:
                raise InvalidManifestException("Missing required phase: {}"
                                               "".format(command))

    def parse_comment(self, bareline):
        text_comment = bareline[1:].strip()
        return "comment", dict(text=text_comment)

    def parse_command(self, bareline):
        words = shlex.split(bareline)
        command = words[0]
        kwargs = {}
        for arg in words[1:]:
            k, v = arg.split("=", 1)
            kwargs[k] = v
        return command, kwargs

    def parse_manifest(self, manifest_path):
        """Return a list of phases from the manifest file."""
        phases = []
        with open(manifest_path) as manifest_file:
            previous_phase = None
            for line in manifest_file:
                bareline = line.strip()
                if not bareline or bareline.startswith("##"):
                    continue

                if bareline.startswith("#"):
                    command, kwargs = self.parse_comment(bareline)
                else:
                    command, kwargs = self.parse_command(bareline)

                config = kwargs.get('config')
                if command == "include" and config:
                    phases.extend(self.parse_manifest(self.spec.get_config(base=config, stage=self.stage)))
                    continue

                try:
                    current_phase = phase.build(command, **kwargs)
                except phase.NoSuchPhaseError as e:
                    raise InvalidManifestException(e.message)

                if (previous_phase and previous_phase.name == "comment" and
                        current_phase.name == "comment"):
                    previous_phase.options['text'] += (
                        "\n" + current_phase.options['text'])
                else:
                    phases.append(current_phase)

                previous_phase = current_phase
        return phases

    def run(self, project, workspace, series, stage=None):
        "Run all phases defined in this manifest"
        auto_repo = True
        if "repo" in [p.name for p in self.phases]:
            # Repo step is being called explicitly - don't call it
            # automatically
            auto_repo = False
        auto_secrets = True
        if "secrets" in [p.name for p in self.phases]:
            # Secrets are being called explicitly - don't call it
            # automatically
            auto_secrets = False
        for phase_name in self.phases:
            if phase_name.name in ["deploy", "build", "script", "verify"]:
                phase_name.run(project, workspace, series, stage,
                               auto_secrets=auto_secrets)
            elif phase_name.name == "collect":
                phase_name.run(project, workspace, series, stage,
                               auto_repo=auto_repo)
            else:
                phase_name.run(project, workspace, series, stage)


#######################################
# Monitor
#######################################

class Monitor(object):

    def __init__(self):
        self.CHANGED = 100

    def run(self, project, workspace, series, stage=None):
        """"Check for new revnos for branches
            Store the previous revnos in monitor dir
        """
        self.spec = workspace.spec
        self.monitor_dir = os.path.join(
            project.project_series, '{}.{}.{}'.format(
                workspace.name, stage, 'monitor'))
        mkdirs(self.monitor_dir)
        new = False
        with open(self.spec.get_config('collect', stage)) as config:
            for line in config:
                previous_revno = None
                if not line.strip().startswith("#") and not len(line.strip()) == 0:
                    options = {}
                    source = line.split()[1]
                    if ";" in source:
                        source, opt_string = source.strip().rsplit(";", 1)
                        for opt in opt_string.split(","):
                            options.update(dict((opt.split("="),)))
                    digest = hashlib.sha1(source).hexdigest()
                    branch = bzr.BzrBranch(source)
                    # Not a branch, could be a directory (@) or http source:
                    # ignore
                    if branch.is_branch:
                        current_revno = branch.revno
                        digestpath = os.path.join(self.monitor_dir, digest)
                        if os.path.exists(digestpath):
                            with open(digestpath) as digestfile:
                                for line in digestfile:
                                    previous_revno = line.split()[0]
                                    logging.debug(
                                        "Read revno={} from {}: {}"
                                        "".format(previous_revno,
                                                  digestpath, line))
                        if not previous_revno == str(current_revno):
                            if "revno" in options:
                                logging.debug("Revno {} for branch {} has "
                                              "changed but configuration "
                                              "calls for revno {}"
                                              "".format(current_revno, source,
                                                        options["revno"]))
                                current_revno = options["revno"]
                            else:
                                new = True
                                logging.info("Revno {} for branch {} has "
                                             "changed from previous revno {}"
                                             "".format(current_revno, source,
                                                       previous_revno))
                            with open(digestpath, 'w') as digestfile:
                                digestfile.write("{} {}\n".format(
                                    current_revno, source))
        logging.info("Cache stats: {}".format(CacheDisk.get_stats()))
        if new:
            logging.info("Changes to branches for project: {}"
                         "".format(project.name))
            sys.exit(self.CHANGED)


#######################################
# Project
#######################################

class DuplicateProjectError(Exception):
    pass


class ProjectCreationError(Exception):
    pass


class ProjectNameError(Exception):
    pass


class ProjectDoesNotExistError(Exception):
    pass


class Project(object):
    "A mojo project"
    token_re = re.compile(r"^[a-zA-Z0-9_.-]+$").search
    token_desc = "{} must only contain alphanumerics, underscores, or periods."

    def __init__(self, name, series, container_class=LXCContainer):
        if not name:
            raise ProjectCreationError("No project name specified. Use "
                                       "--project or set MOJO_PROJECT")
        if not bool(self.token_re(name)):
            raise ProjectNameError(self.token_desc.format("Project names"))
        if name in RESERVED_PROJECT_NAMES:
            raise ProjectNameError("Projects may not be named {}".format(name))
        self.name = name
        self.series = series
        self.owner = os.environ.get('SUDO_USER')
        if not self.owner:
            # map my running uid to username
            self.owner = pwd.getpwuid(os.getuid()).pw_name

        # ### Paths ###
        # /srv/mojo/<prj>
        self.root = Project.root_path(self.name)
        # /srv/mojo/<prj>/<series>
        self.project_series = os.path.join(self.root, self.series)

        # Bootstrap container information for pre-existing projects
        # All NoContainer projects will have a project info file
        # If project info does not exist it is safe to assume the
        # default container_class of LXCContainer
        self.project_info = self.get_project_info()
        if self.project_info.get('container_class'):
            logging.info("Using pre-existing container class {}"
                         "".format(self.project_info.get('container_class')))
            if self.project_info.get('container_class') == 'lxc':
                container_class = LXCContainer
            elif self.project_info.get('container_class') == 'containerless':
                container_class = NoContainer
            # else use container_class passed to init
            # this defaults to container_class=LXCContainer
        self.container = container_class(self.name, self.series)
        if self.container.container_class == 'lxc':
            # /srv/mojo/<prj>/<series>/ROOTFS
            self.project_container_root = os.path.join(self.project_series, "ROOTFS")
            # /srv/mojo/<prj>/<series>/ROOTFS/srv/mojo/<prj>/<series>
            self.container_project_series = self.project_container_root + self.project_series
            if self.project_info.get('container_root'):
                self.container.container_root = self.project_info.get('container_root')
            # If the lxc does not already exist and lxc >= 1.0.0
            # use location setting
            # Can't use lxc location on precise's version of lxc missing
            # -B --dir option
            elif not self.container.exists() and \
                    self.container.validate_lxc_version():
                self.container.container_root = self.project_container_root
        elif self.container.container_class == 'containerless':
            self.container.container_root = self.project_series
            self.project_container_root = self.project_series
            self.container_project_series = self.project_series

    def get_project_info(self):
        """Get pre-existing project information from the .project file.
        Return a dictionary of pre-existing information."""
        info_file = os.path.join(self.project_series, '.project')
        if os.path.exists(info_file):
            with open(info_file, 'r') as info:
                return json.loads(info.read())
        else:
            return {}

    def write_project_info(self):
        """Create a project information file. This will allow for state
        to be learned for subsequent mojo commands without requiring prior
        knowledge of the pre-existing project or command line switches"""
        info_file = os.path.join(self.project_series, '.project')
        # This can be expanded down the road with more data
        project_info = {'container_class': self.container.container_class,
                        'container_root': self.container.container_root}
        with open(info_file, 'w') as info:
            info.write(json.dumps(project_info))

    @classmethod
    def root_path(cls, name):
        if not bool(cls.token_re(name)):
            raise ProjectNameError(cls.token_desc.format("Project names"))
        if name in RESERVED_PROJECT_NAMES:
            raise ProjectNameError("Projects may not be named {}".format(name))

        return os.path.join(MOJO_ROOT, name)

    @classmethod
    def list_series(cls, name):
        if not bool(cls.token_re(name)):
            raise ProjectNameError(cls.token_desc.format("Project names"))
        if name in RESERVED_PROJECT_NAMES:
            raise ProjectNameError("Projects may not be named {}".format(name))

        try:
            return [s for s in sorted(os.listdir(cls.root_path(name)))]
        except OSError as ose:
            if ose.errno == errno.ENOENT:
                raise ProjectDoesNotExistError(
                    'No such project: {}'.format(name))
            raise

    @classmethod
    def list(cls, only_names):
        "List all Mojo projects"
        proj_list = []
        for proj in os.listdir(MOJO_ROOT):
            if proj in RESERVED_PROJECT_NAMES:
                continue
            # Ignore any directories we don't have permissions to read in case
            # we have a multi-tenant mojo install. This is not expected as a
            # long term solution to multi-tenant mojo installs, as we should be
            # changing MOJO_ROOT for that, but it's not currently supported.
            try:
                os.listdir(os.path.join(MOJO_ROOT, proj))
            except OSError:
                continue
            if not only_names:
                series = " ".join(cls.list_series(proj))
                proj = "{}: {}".format(proj, series)
            proj_list.append(proj)
        return proj_list

    @needs_root
    def create(self):
        "Create a new container and project"
        logging.info("Checking {} project path at {} ...".format(
            self.name, self.root))
        if os.path.exists(self.root):
            logging.warning("Project {} already exists at {}. Using existing "
                            "project".format(self.name, self.root))
        else:
            try:
                mkdirs(self.root, self.owner, mode=0700)
            except OSError as e:
                raise ProjectCreationError(
                    "Could not create project root: {}".format(e.strerror))

        if os.path.exists(self.project_series):
            logging.warning("Project {} project_series already exists at {}, "
                            "using it.".format(self.name, self.project_series))
        else:
            try:
                mkdirs(self.project_series, self.owner, mode=0700)
            except OSError as e:
                raise ProjectCreationError(
                    "Could not create project project_series: "
                    "{}".format(e.strerror))

        logging.info("Checking {} project container at {} ..."
                     "".format(self.name, self.container.conf))
        if not self.container.exists():
            logging.info("Not found, creating container")
            self.container.create()

        logging.info("Checking {} project container is setup ..."
                     "".format(self.name))
        if not self.container.is_setup():
            logging.info("Not setup - setting up container for mojo")
            self.container.setup()

        if self.container.container_class != 'containerless':
            if self.project_container_root != self.container.container_root:
                logging.info("Checking {} project container symlink {} -> {} ..."
                             "".format(self.name, self.container.container_root, self.project_container_root))
                if not os.path.exists(self.project_container_root):
                    os.symlink(self.container.container_root, self.project_container_root)

            logging.info("Checking {} project dir {} ..."
                         "".format(self.name, self.container_project_series))
            if not os.path.exists(self.container_project_series):
                mkdirs(self.container_project_series, self.owner, mode=0700)

            sudoers_file = os.path.join(self.project_container_root, "etc", "sudoers.d",
                                        "ubuntu-all")
            logging.info("Writing {} sudoers file: {}".format(
                         self.name, sudoers_file))
            with open(sudoers_file, "w") as f:
                f.write("ubuntu ALL=(ALL) NOPASSWD: ALL\n")
            os.chmod(sudoers_file, 0400)

        if not os.path.exists(os.path.join(self.project_series, '.project')):
            self.write_project_info()

    def delete(self, force=True):
        """Delete a project and its LXC container - if no series remain,
        also delete project root"""
        try:
            self.container.destroy()
        except NoSuchContainerError:
            if not force:
                raise
        try:
            shutil.rmtree(self.project_series)
        except OSError:
            if not force:
                raise

        # If no series remain, delete project root
        if not Project.list_series(self.name):
            logging.info("Destroying project root: {}"
                         "".format(Project.root_path(self.name)))
            try:
                shutil.rmtree(Project.root_path(self.name))
            except OSError:
                if not force:
                    raise

    def ws_link(self, title):
        return os.path.join(self.project_series, str(title))

    def create_workspace(self, spec_url, title, stage=None,
                         manifest_file='manifest', clean_spec=True,
                         logfile=None):

        # Validate the project exists
        if self.name not in self.list(True):
            print "Project, '{}', does not appear to exist.".format(self.name)
            print "Please run 'sudo mojo project-new --series {} {}' to " \
                  "create the project.".format(self.series, self.name)
            print "Or check the file permissions on {}".format(self.root)
            sys.exit(1)
        if not self.container.exists() \
                or not self.container.is_setup():
            print "The container project and series path for {} does not appear " \
                  "to exist at {}.".format(self.name, self.container_project_series)
            print "Check to make sure the project and series exist and " \
                  "permissions are set properly."
            print "Please run 'sudo mojo project-new --series {} {}' to " \
                  "create the project if needed.".format(self.series, self.name)
            print "Or check the file permissions on {}".format(self.container_project_series)
            sys.exit(1)

        # If we're using nested directories for specs, make sure our workspace
        # symlink doesn't need additional sub-directories created.
        if "/" in title:
            title = title.replace("/", "-")
        ws_path = os.path.join(self.container_project_series, str(title))
        mkdirs(ws_path, self.owner)
        if self.container.container_class == 'containerless':
            mkdirs(self.ws_link(title), self.owner)
        else:
            mksymlink(ws_path, self.ws_link(title))
        ws = Workspace(title, self.ws_link(title), stage=stage,
                       manifest_file=manifest_file, logfile=logfile)
        ws.prepare()
        spec = Spec(ws.spec_dir, spec_url, stage=stage, manifest_file=manifest_file)
        spec.initialize(clean_spec)
        return ws

    def workspace(self, workspace_name, logfile=None):
        return Workspace(workspace_name, self.ws_link(workspace_name),
                         logfile=logfile)

    def delete_workspace(self, workspace_name):
        ws = self.workspace(workspace_name)
        ws.delete()
        if os.path.lexists(self.ws_link(workspace_name)):
            os.unlink(self.ws_link(workspace_name))

    def workspace_list(self):
        workspaces = [w for w in os.listdir(self.container_project_series)
                      if w not in ['LOCAL', '.project']]
        workspace_output = ["Workspace: Spec URL",
                            "==================="]
        for workspace in workspaces:
            spec = self.workspace(workspace).spec
            if spec:
                workspace_output.append("{}: {}"
                                        "".format(workspace,
                                                  spec.remote_branch.location))
            else:
                logging.warn("Workspace directory {} has no valid spec"
                             "".format(workspace))
        return workspace_output

    @property
    def utils(self):
        return self.container.utils


#######################################
# Workspace
#######################################

class NoSuchWorkspaceError(Exception):
    pass


class Workspace(object):
    "A mojo workspace"
    def __init__(self, name, path, stage=None, manifest_file=None, logfile=None):
        self.name = name
        self.manifest_file = manifest_file
        if not path:
            self.path = os.path.abspath(os.curdir)
            logging.info("No workspace path specified, using {}"
                         "".format(self.path))
        else:
            self.path = os.path.abspath(path)
        if not os.path.exists(self.path):
            raise NoSuchWorkspaceError(path)
        self.stage = stage
        self.logfile = logfile
        self.create_file_logger()

    @property
    def spec(self):
        if os.path.exists(self.spec_dir):
            spec_repo = bzr.BzrBranch(self.spec_dir)
            if spec_repo.is_branch:
                return Spec(self.spec_dir, stage=self.stage,
                            manifest_file=self.manifest_file)
        return None

    @property
    def all_dirs(self):
        return [self.path,
                self.build_dir,
                self.local_dir,
                self.log_dir,
                self.spec_dir,
                ]

    def prepare(self):
        "Set up the workspace. Returns True on success, False on failure."
        for path in self.all_dirs:
            if not os.path.isdir(path):
                try:
                    os.makedirs(path)
                except OSError as e:
                    logging.error("{} ({})".format(e.strerror, e.filename))
                    return False
        return True

    def clean(self):
        "Reset the workspace"
        map(shutil.rmtree, (d for d in self.all_dirs if os.path.exists(d)))
        return self.prepare()

    def delete(self):
        "Destroy the workspace"
        realpath = os.path.realpath(self.path)
        shutil.rmtree(realpath)

    def create_file_logger(self):
        "If we haven't already, attach a file logger"
        if logging.getLogger('file').handlers:
            return

        if not self.logfile:
            self.logfile = os.path.join(self.log_dir, "mojo.log")

        try:
            logFormatter = logging.Formatter(
                fmt="%(asctime)s [%(levelname)s] %(message)s",
                datefmt="%Y-%m-%d %H:%M:%S")

            rootLogger = logging.getLogger()
            fileLogger = logging.getLogger('file')
            # If we send output to the file logger specifically, don't propagate it
            # to the root logger as well to avoid duplicate output. So if we want
            # to only send logging output to the file, you would do this:
            #  logging.getLogger('file').info("message for logfile only")
            # rather than this:
            #  logging.info("message for console and logfile")
            fileLogger.propagate = False

            fileHandler = logging.FileHandler(self.logfile)
            fileHandler.setFormatter(logFormatter)
            rootLogger.addHandler(fileHandler)
            fileLogger.addHandler(fileHandler)
        except IOError:
            logging.error("Unable to write to logfile: {}".format(self.logfile))

    @property
    def build_dir(self):
        return os.path.join(self.path, "build")

    @property
    def repo_dir(self):
        return os.path.join(self.path, "charms")

    @property
    def local_dir(self):
        return os.path.join(self.path, "local")

    @property
    def log_dir(self):
        if self.logfile:
            return os.path.dirname(os.path.abspath(self.logfile))
        return os.path.join(self.path, "log")

    @property
    def spec_dir(self):
        return os.path.join(self.path, "spec")
