# Copyright 2014, 2015 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
from __future__ import print_function
import argparse
import logging
import subprocess
import sys
import os
import pkg_resources

import mojo

from .contain import (
    LXCContainer,
    NoContainer,
    NoSuchContainerError
)


class ArgumentError(Exception):
    pass


def action(name, args):
    project = get_project(args)
    workspace = get_workspace(args)
    stage = get_stage(args)
    series = get_series(args)
    logfile = get_logfile(args)
    options = args.options
    os.environ["MOJO_PROJECT"] = project
    os.environ["MOJO_WORKSPACE"] = workspace
    workspace = mojo.Project(project, series).workspace(workspace, logfile)
    project = mojo.Project(project, series)
    phase = mojo.phase.build(name, **options)
    phase.run(project, workspace, series, stage)


class OptionParser(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        try:
            options = dict(option.split("=", 1) for option in values)
        except ValueError:
            print("Options must be key=value pairs.", file=sys.stderr)
            parser.print_help(file=sys.stderr)
            sys.exit(1)
        setattr(namespace, self.dest, options)


# -----------------------
# helpers
# -----------------------

def get_series(args):
    """Return the series from args or the environment.

    Raises ArgumentError if no series can be determined.
    """
    series = args.series or os.environ.get("MOJO_SERIES")
    if not series:
        raise ArgumentError("Series setting missing. Use --series or set "
                            "MOJO_SERIES")
    return series


def get_workspace(args):
    """Return the workspace from args or the environment.

    Raises ArgumentError if no workspace can be determined.
    """
    workspace = args.workspace or os.environ.get("MOJO_WORKSPACE")
    if not workspace:
        raise ArgumentError("Workspace setting missing. Use --workspace or "
                            "set MOJO_WORKSPACE")
    return workspace


def get_project(args, use_spec=False):
    """Return the project from args or the environment.

    Priority for setting project: command line switch then
    MOJO_PROJECT environment variable

    If use_spec is true and the above priority settings are
    missing, then the project will be guessed from spec_url.
    A warning will be printed to this effect.

    Raises ArgumentError if no project can be determined.
    """

    # command line switch, --project
    project = args.project
    # MOJO_PROJECT environment variable
    if not project:
        project = os.environ.get("MOJO_PROJECT")
    # Guess based on spec_url and warn
    if not project and use_spec:
        spec_url, workspace = get_spec_url_and_workspace(args)
        project = spec_url.rsplit("/")[-1]
        if not project:
            project = args.spec_url.rsplit("/")[-2]
        logging.warn("Guessing the project is {}. Use --project or set "
                     "MOJO_PROJECT to be more explicit.".format(project))
    # Fail gracefully
    if not project:
        raise ArgumentError("Project setting missing. Use --project or set "
                            "MOJO_PROJECT")
    return project


def get_stage(args):
    """Return the stage from args or the environment.

    Returns 'devel' by default.
    """
    return args.stage or os.environ.get("MOJO_STAGE") or 'devel'


def get_spec_url_and_workspace(args):
    """Return the spec_url and workspace from args or the
    environment.

    Returns spec_url and workspace.

    Raises ArgumentError if either spec_url or workspace
    cannot be determined.
    """
    spec_url = args.spec_url or os.environ.get("MOJO_SPEC")
    workspace = args.workspace or os.environ.get("MOJO_WORKSPACE")

    if not spec_url or not workspace or spec_url == workspace:
        raise ArgumentError("Missing spec_url and/or workspace. Set "
                            "MOJO_SPEC and/or MOJO_WORKSPACE")

    return (spec_url, workspace)


def get_logfile(args):
    "Return logfile from args or the environment"
    return args.logfile or os.environ.get("MOJO_LOGFILE")


def get_manifest_file(args):
    """Return the manifest file from args or default.

    We can't depend on default value of --manifest-file as it's
    only present for the run subcommand.
    """
    if hasattr(args, 'manifest'):
        return args.manifest

    return 'manifest'


def get_container(args):
    """Return the container from args or the environment.
    Returns 'LXCContainer' by default.
    Currently this is only required at project creation time.
    Stateful project information will be gathered on subsequent
    mojo commands. If this behavior changes, the defaults below
    may need to change.
    """

    if args.container:
        container = args.container
    elif os.environ.get("MOJO_CONTAINER"):
        container = os.environ.get("MOJO_CONTAINER")
    else:
        # For now default to lxc
        container = 'lxc'

    if container == 'lxc':
        return LXCContainer
    elif container == 'containerless':
        return NoContainer
    else:
        raise NoSuchContainerError("{} is an unsupported container type."
                                   "".format(container))


# -----------------------
# commands
# -----------------------
def create_project(args):
    "Set up a Mojo project and LXC container"
    project_name = args.name
    series = get_series(args)
    project = mojo.Project(project_name, series,
                           container_class=get_container(args))
    project.create()


def destroy_project(args):
    try:
        if args.series:
            series_list = (args.series,)
        else:
            series_list = mojo.Project.list_series(args.name)

        projects = [mojo.Project(args.name, series) for series in series_list]

        for project in projects:
            logging.info("Destroying project: {}, series: {}"
                         "".format(project.name, project.series))
            project.delete()

    except mojo.ProjectNameError:
        logging.error('Invalid project name: {}'.format(args.name))
        sys.exit(1)
    except mojo.ProjectDoesNotExistError:
        logging.error('Invalid project: {}'.format(args.name))
        sys.exit(1)
    except mojo.ProjectCreationError:
        logging.exception('Invalid project: {}'.format(args.name))
        sys.exit(1)


def list_projects(args):
    map(print, mojo.Project.list(only_names=args.names))


def create_workspace(args):
    "Set up workspace and fetch a copy of the spec"
    project_name = get_project(args)
    series = get_series(args)
    project = mojo.Project(project_name, series)
    stage = get_stage(args)
    manifest_file = get_manifest_file(args)
    logfile = get_logfile(args)
    spec_url, workspace = get_spec_url_and_workspace(args)

    try:
        ws = project.create_workspace(spec_url, workspace,
                                      stage=stage,
                                      manifest_file=manifest_file,
                                      logfile=logfile)
    except OSError as e:
        print("Are you sure the project exists?", file=sys.stderr)
        if args.blunt:
            raise
        sys.exit(2)
    except AttributeError as e:
        print("Unable to branch {}".format(args.spec_url))
        print("Please check the branch exists and you have permissions to "
              "access it")
        # Now we need to clean up the workspace
        mojo.Project(project_name, series).delete_workspace(workspace)
        sys.exit(2)
    logging.info("Workspace created. Validating manifest")
    try:
        manifest = ws.spec.manifest
        logging.info("Manifest OK ({})".format(manifest.config))
    except mojo.ConfigNotFoundException as e:
        logging.error(e.message)
        sys.exit(1)
    except ValueError as e:
        logging.error(e.message)
        sys.exit(1)


def destroy_workspace(args):
    series = get_series(args)
    project_name = get_project(args)

    try:
        mojo.Project(project_name, series).workspace(args.workspace)
    except mojo.NoSuchWorkspaceError:
        logging.error(
            'No such workspace (project: {project}; series: {series}): {workspace}'.format(
                project=project_name,
                series=series,
                workspace=args.workspace,
            ))
        sys.exit(1)
    except mojo.ProjectNameError:
        logging.error('Invalid project name: {}'.format(project_name))
        sys.exit(1)
    except mojo.ProjectDoesNotExistError:
        logging.error('Invalid project: {}'.format(project_name))
        sys.exit(1)
    except mojo.ProjectCreationError:
        logging.exception('Invalid project: {}'.format(project_name))
        sys.exit(1)

    mojo.Project(project_name, series).delete_workspace(args.workspace)


def list_workspaces(args):
    mojo_project = get_project(args)
    mojo_series = get_series(args)
    project = mojo.Project(mojo_project, mojo_series)
    map(print, project.workspace_list())


def display_help(args):
    "Display help for sub-commands"
    if not args.command:
        args.parser.print_help()
    else:
        args.parser.parse_args([args.command, '-h'])
    sys.exit(0)


def show_deploy(args):
    "Show rendered juju-deployer configuration"
    project = get_project(args)
    workspace = get_workspace(args)
    stage = get_stage(args)
    series = get_series(args)
    options = args.options
    workspace = mojo.Project(project, series).workspace(workspace)
    project = mojo.Project(project, series)
    phase = mojo.phase.build('deploy', **options)
    phase.show(project, workspace, series, stage)


def diff_deploy(args):
    """
    Show juju-deployer diff. Generate a delta between a
    configured deployment and a running environment
    """
    project = get_project(args)
    workspace = get_workspace(args)
    stage = get_stage(args)
    series = get_series(args)
    options = args.options
    workspace = mojo.Project(project, series).workspace(workspace)
    project = mojo.Project(project, series)
    phase = mojo.phase.build('deploy', **options)
    phase.diff(project, workspace, series, stage)


def builddeps(args):
    "Install build dependencies into project container"
    action('builddeps', args)


def collect(args):
    "Collect build resources using config-manager"
    action('collect', args)


def build(args):
    "Run deployment's build script"
    action('build', args)


def script(args):
    "Run deployment's build script"
    action('script', args)


def repo(args):
    "Build a charm repository"
    action('repo', args)


def sleep(args):
    "Sleep"
    action('sleep', args)


def secrets(args):
    "Secrets"
    action('secrets', args)


def deploy(args):
    "Build or update services"
    action('deploy', args)


def volumes(args):
    "Mount volumes"
    action('volumes', args)


def verify(args):
    "Verify the running service"
    action('verify', args)


def run_from_manifest(args):
    project_name = get_project(args, use_spec=True)
    series = get_series(args)
    stage = get_stage(args)
    logfile = get_logfile(args)
    spec_url, workspace = get_spec_url_and_workspace(args)
    if args.keep_local_spec_changes:
        clean_spec = False
    else:
        clean_spec = True
    project = mojo.Project(project_name, series)
    workspace = project.create_workspace(spec_url, workspace,
                                         stage=stage,
                                         manifest_file=args.manifest_file,
                                         clean_spec=clean_spec,
                                         logfile=logfile)
    manifest = workspace.spec.manifest
    manifest.run(project, workspace, series, stage)


def monitor_specification(args):
    project_name = get_project(args, use_spec=True)
    project = mojo.Project(project_name, args.series)
    workspace = project.create_workspace(args.spec_url, args.name,
                                         stage=args.stage,
                                         manifest_file=args.manifest_file)
    monitor = mojo.Monitor()
    monitor.run(project, workspace, args.series, args.stage)


def create_arg_parser():
    # main program parser
    ap = argparse.ArgumentParser()
    ap.add_argument("--blunt", action="store_true",
                    help="Handle fewer exceptions")
    ap.add_argument("--log_level", "-l", default='INFO',
                    help="Log level. Defaults to INFO")
    ap.add_argument("--logfile", "-L", default="",
                    help="File to log to in addition to stdout. Defaults to "
                    "MOJO_WORKSPACE/log/mojo.log")
    ap.add_argument('--version', action='version',
                    version=pkg_resources.require("mojo")[0].version)

    # subcommand parser
    sp = ap.add_subparsers(title="subcommands", dest="subparser_name")

    # options parser for phase commands
    phaseopts = argparse.ArgumentParser(add_help=False)
    phaseopts.add_argument("--stage", help="Deployment stage (default: devel)")
    phaseopts.add_argument("--options",
                           help="Phase options. i.e. config=CONFIG_FILENAME",
                           nargs="*", metavar="key=value", action=OptionParser,
                           default={})
    phaseopts.add_argument("-p", "--project",
                           help="Project name. Overrides the MOJO_PROJECT "
                           "environment variable.")
    phaseopts.add_argument("-w", "--workspace", help="Workspace name. "
                           "Overrides the MOJO_WORKSPACE environment variable.")
    phaseopts.add_argument("-s", "--series", help="Distro series. Overrides"
                           "the MOJO_SERIES environment variable.")

    # project-new subcommand - creates a new mojo project
    prjnewcmd = sp.add_parser("project-new",
                              description="Create a new Mojo project and "
                              "container")
    prjnewcmd.add_argument("-s", "--series", help="Distro series",
                           required=False)
    prjnewcmd.add_argument("name",
                           help="Base name of the project ([a-zA-Z0-9._]*)")
    prjnewcmd.add_argument('-c', "--container", default="lxc",
                           choices=['lxc', 'containerless'],
                           help="Container. Options 'lxc' or 'containerless'. "
                                "Defaults to 'lxc'")
    prjnewcmd.set_defaults(func=create_project)

    # project-destroy subcommand - deletes a mojo project
    prjdelcmd = sp.add_parser("project-destroy",
                              description="Destroy a Mojo project and "
                              "container")
    prjdelcmd.add_argument("name",
                           help="Base name of the project ([a-zA-Z0-9._]*)")
    prjdelcmd.add_argument("-s", "--series", help="Distro series "
                           "(default=<all found>)")
    prjdelcmd.set_defaults(func=destroy_project)

    # project-list
    prjlistcmd = sp.add_parser("project-list",
                               description="List existing projects")
    prjlistcmd.add_argument("-n", "--names", action="store_true",
                            help="List only project names, without series")
    prjlistcmd.set_defaults(func=list_projects)

    # workspace-new subcommand - creates a new workspace for a specific spec
    # revision
    wsnewcmd = sp.add_parser("workspace-new",
                             description="Initialize a mojo workspace")
    wsnewcmd.add_argument("spec_url", nargs='?', default=None,
                          help="Bzr URL to the Mojo spec. Overrides the "
                               "MOJO_SPEC environment variable.")
    wsnewcmd.add_argument("workspace", nargs='?', default=None,
                          help="Name of the workspace (suggested: revno of "
                               "spec). Overrides the MOJO_WORKSPACE "
                               "environment variable. If workspace is "
                               "specified spec_url must be also.")
    wsnewcmd.add_argument("-p", "--project",
                          help="Project name (equivalent to MOJO_PROJECT "
                          "environment variable)")
    wsnewcmd.add_argument("-s", "--series", help="Distro series. Overrides"
                          "the MOJO_SERIES environment variable.")
    wsnewcmd.add_argument("--stage",
                          help="Deployment stage. Defaults to 'devel'")
    wsnewcmd.set_defaults(func=create_workspace, parser=wsnewcmd)

    # workspace-destroy subcommand - deletes a mojo project
    wsdelcmd = sp.add_parser("workspace-destroy",
                             description="Destroy a mojo workspace")
    wsdelcmd.add_argument("-p", "--project",
                          help="Project name (equivalent to MOJO_PROJECT "
                          "environment variable)")
    wsdelcmd.add_argument("-s", "--series", help="Distro series. Overrides "
                          "the MOJO_SERIES environment variable.")
    wsdelcmd.add_argument("workspace",
                          help="Workspace name. Required "
                          "option. Will not use MOJO_WORKSPACE environment "
                          "variable by design.")
    wsdelcmd.set_defaults(func=destroy_workspace)

    # workspace-list subcommand
    wslistcmd = sp.add_parser("workspace-list",
                              description="List workspaces for a project")
    wslistcmd.add_argument("-p", "--project",
                           help="Project name (equivalent to MOJO_PROJECT "
                           "environment variable)")
    wslistcmd.add_argument("-s", "--series", help="Distro series. Overrides"
                           "the MOJO_SERIES environment variable.")
    wslistcmd.set_defaults(func=list_workspaces)

    # builddeps subcommand - install build dependencies for project
    depscmd = sp.add_parser("builddeps", description="builddeps",
                            parents=[phaseopts])
    depscmd.set_defaults(func=builddeps)

    # collect subcommand - collects code branches to build directory
    clctcmd = sp.add_parser("collect",
                            description="Collect resources for build",
                            parents=[phaseopts])
    clctcmd.set_defaults(func=collect)

    # build subcommand - run project build script inside container
    buildcmd = sp.add_parser("build",
                             description="Execute project build script",
                             parents=[phaseopts])
    buildcmd.set_defaults(func=build)

    # repo subcommand - assembles charms into a repository
    repocmd = sp.add_parser("repo", description="Assemble a charm repository",
                            parents=[phaseopts])
    repocmd.set_defaults(func=repo)

    # sleep subcommand - sleeps for n seconds
    sleepcmd = sp.add_parser("sleep", description="Sleep for time in seconds",
                             parents=[phaseopts])
    sleepcmd.set_defaults(func=sleep)

    # secrets subcommand - secretss for n seconds
    secretscmd = sp.add_parser("secrets",
                               description="Copy secrets from staged "
                               "location to workspace local",
                               parents=[phaseopts])
    secretscmd.set_defaults(func=secrets)

    # script subcommand - run an arbitrary script inside container
    scriptcmd = sp.add_parser("script", description="Execute a project script",
                              parents=[phaseopts])
    scriptcmd.set_defaults(func=script)

    # deploy subcommand - run a juju-deployer session
    deploycmd = sp.add_parser("deploy",
                              description="Deploy services and/or create "
                              "relations",
                              parents=[phaseopts])
    deploycmd.add_argument("--ignore", action="store_true", default=False,
                           help="Exit 0 even if service deployment fails")
    deploycmd.set_defaults(func=deploy)

    # deploy-show subcommand
    deployshowcmd = sp.add_parser("deploy-show",
                                  description="Show rendered juju-deployer "
                                              "configuration",
                                  parents=[phaseopts])
    deployshowcmd.set_defaults(func=show_deploy)

    # deploy-diff subcommand
    deploydiffcmd = sp.add_parser("deploy-diff",
                                  description="Show juju-deployer diff. "
                                              "Generate a delta between a "
                                              "configured deployment and a "
                                              "running environment.",

                                  parents=[phaseopts])
    deploydiffcmd.set_defaults(func=diff_deploy)

    # volumes subcommand - mount volumes
    volcmd = sp.add_parser("volumes", description="Mount volumes",
                           parents=[phaseopts])
    volcmd.set_defaults(func=volumes)

    # verify subcommand - verify that a deployment was successful
    verifycmd = sp.add_parser("verify",
                              description="Verify that the deployment is "
                              "working",
                              parents=[phaseopts])
    verifycmd.set_defaults(func=verify)

    # run subcommand - run an entire spec per its manifest
    runcmd = sp.add_parser("run",
                           description="Run an entire deployment from start "
                           "to finish")
    runcmd.add_argument("spec_url", nargs='?', default=None,
                        help="Bzr URL to the Mojo spec. Overrides the "
                             "MOJO_SPEC environment variable.")
    runcmd.add_argument("workspace", nargs='?', default=None,
                        help="Name of the workspace (suggested: revno of spec)"
                        ". Overrides the MOJO_WORKSPACE environment variable. "
                        "If workspace is specified spec_url must be also.)")
    runcmd.add_argument("--manifest-file", default="manifest",
                        help="The particular manifest file to run")
    runcmd.add_argument("--series", help="Distro series. Overrides"
                        "the MOJO_SERIES environment variable.")
    runcmd.add_argument("--stage",
                        help="Deployment stage. Defaults to 'devel'")
    runcmd.add_argument("--project",
                        help="Name for the project (defaults to spec name)")
    runcmd.add_argument("-k", "--keep-local-spec-changes", help="Keep "
                        "local edits. This option allows you to make local "
                        "edits in $MOJO_SPEC_DIR for quick testing",
                        action="store_true")
    runcmd.set_defaults(func=run_from_manifest)

    # monitor subcommand - monitor specification collect file for
    # updates
    monitorcmd = sp.add_parser("monitor",
                               description="Monitor specification collect "
                               "file for updates")
    monitorcmd.add_argument("spec_url", help="Bzr URL to the Mojo spec")
    monitorcmd.add_argument("name",
                            help="Name of the workspace (suggested: revno)")
    monitorcmd.add_argument("--stage",
                            help="Deployment stage. Defaults to 'devel'")
    monitorcmd.add_argument("--project",
                            help="Name for the project (defaults to spec "
                            "name)")
    monitorcmd.add_argument("--series", help="Distro series. Overrides"
                            "the MOJO_SERIES environment variable.")
    monitorcmd.add_argument("--cache_ttl", type=int, default=3600,
                            help="Number of seconds to cache bzr revnos"
                            "(default=%s)" % 3600)
    monitorcmd.set_defaults(func=monitor_specification)

    # help subcommand - get help for other subcommands
    helpcmd = sp.add_parser("help", description="Interactive help")
    helpcmd.add_argument("command", nargs='?', default=None,
                         help="Get help for this sub-command.")
    helpcmd.set_defaults(func=display_help, parser=ap)

    return ap


def run_with_args(args, ap=None):
    if hasattr(args, "cache_ttl") and args.cache_ttl is not None:
        mojo.CacheDisk.ttl(args.cache_ttl)
    try:
        args.func(args)
    except ArgumentError as e:
        logging.error(e.message)
        if ap is not None:
            ap.parse_args([args.subparser_name, '-h'])
        sys.exit(1)
    except mojo.utils.Unprivileged:
        logging.error("This sub-command, {}, must be run as root"
                      "".format(args.subparser_name))
        sys.exit(1)
    except mojo.phase.CollectPhaseException as e:
        logging.error(e)
        sys.exit(1)
    except mojo.ConfigNotFoundException as e:
        logging.error(e.message)
        print("Searched:", file=sys.stderr)
        for path in e.search_list:
            print("  {}".format(path), file=sys.stderr)
        sys.exit(1)
    except (mojo.InvalidManifestException,
            mojo.charm_repo.DuplicateCharmError,
            mojo.phase.FileDoesNotExistException,
            mojo.phase.InvalidStopOnReturnCodeException,
            mojo.phase.InvalidYAMLException) as e:
        logging.error(e.message)
        sys.exit(1)
    except subprocess.CalledProcessError as e:
        logging.info(e.output)
        logging.exception("Error calling external process : return code {}".format(e.returncode))
        sys.exit(1)
    except Exception as e:
        logging.exception("Unknown error")
        sys.exit(1)


# -----------------------
# main
# -----------------------


def main():
    ap = create_arg_parser()
    args = ap.parse_args()
    logFormatter = logging.Formatter(
        fmt="%(asctime)s [%(levelname)s] %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S")
    rootLogger = logging.getLogger()
    rootLogger.setLevel(args.log_level)

    consoleHandler = logging.StreamHandler()
    consoleHandler.setFormatter(logFormatter)
    rootLogger.addHandler(consoleHandler)

    run_with_args(args, ap)
