# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
from contextlib import contextmanager
import mock
import os
from unittest import TestCase
from shutil import rmtree
from tempfile import mkdtemp

import mojo
from mojo.cli import (
    create_arg_parser,
    run_with_args,
)
from mojo.tests.test_bzr import create_unique_branch


@contextmanager
def use_env(**mapping):
    _environ = dict(os.environ)
    os.environ.update(mapping)
    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(_environ)


class ArgParserTestCase(TestCase):
    """The argument parser includes the expected args."""

    def test_arg_parser(self):
        parser = create_arg_parser()

        args = parser.parse_args(
            "workspace-new --project my-project -s trusty "
            "--stage devel "
            "/home/me/my/spec_url my-workspace".split())

        self.assertEqual(args.workspace, 'my-workspace')
        self.assertEqual(args.project, "my-project")
        self.assertEqual(args.series, "trusty")
        self.assertEqual(args.spec_url, "/home/me/my/spec_url")
        self.assertEqual(args.stage, "devel")


class WorkspaceNewTestCase(TestCase):

    def patch_mojo_root(self):
        fake_mojo_root = mkdtemp()
        self.addCleanup(rmtree, fake_mojo_root)
        patcher = mock.patch.object(mojo, 'MOJO_ROOT', fake_mojo_root)
        self.addCleanup(patcher.stop)
        patcher.start()
        return fake_mojo_root

    def create_spec_dir(self):
        spec_dir = mkdtemp()
        self.addCleanup(rmtree, spec_dir)
        create_unique_branch(spec_dir, ('services.yml',))
        return spec_dir

    @mock.patch('mojo.LXCContainer.is_setup', return_value=True)
    @mock.patch('mojo.LXCContainer.exists', return_value=True)
    def test_workspace_new(self, exists, is_setup):
        mojo_root = self.patch_mojo_root()
        spec_dir = self.create_spec_dir()
        parser = create_arg_parser()
        args = parser.parse_args([
            "workspace-new",
            "--project",
            "my-project",
            "-s",
            "trusty",
            spec_dir,
            "my-workspace",
        ])

        # Mock the LXC directory
        lxc_project_series = os.path.join(mojo_root, 'my-project', 'trusty',
                                          'ROOTFS', mojo_root[1:],
                                          'my-project', 'trusty')
        mojo.utils.mkdirs(lxc_project_series)
        setup_file = os.path.join(mojo_root, 'my-project', 'trusty', 'ROOTFS',
                                  'etc', 'init', 'setup-stopped.conf')
        mojo.utils.mkdirs(os.path.dirname(setup_file))
        open(setup_file, 'a').close()

        # Mock Project.list()
        with mock.patch.object(mojo.Project, 'list',
                               return_value='my-project'):
            run_with_args(args)

        expected_dirs = ['build', 'local', 'log', 'spec']
        workspace_dir = os.path.join(mojo_root, 'my-project', 'trusty',
                                     'my-workspace')
        for expected_dir in expected_dirs:
            expected_path = os.path.join(workspace_dir, expected_dir)
            self.assertTrue(os.path.exists(expected_path))

    @mock.patch('mojo.LXCContainer.is_setup', return_value=True)
    @mock.patch('mojo.LXCContainer.exists', return_value=True)
    @mock.patch('mojo.Project.list', return_value='my-project')
    def test_workspace_new_precise(self, project_list, exists, is_setup):
        mojo_root = self.patch_mojo_root()
        spec_dir = self.create_spec_dir()
        parser = create_arg_parser()
        args = parser.parse_args([
            "workspace-new",
            "--project",
            "my-project",
            "-s",
            "precise",
            spec_dir,
            "my-workspace",
        ])

        # Mock the LXC directory
        lxc_project_series = os.path.join(mojo_root, 'my-project', 'precise',
                                          'ROOTFS', mojo_root[1:],
                                          'my-project', 'precise')
        mojo.utils.mkdirs(lxc_project_series)
        setup_file = os.path.join(mojo_root, 'my-project', 'precise', 'ROOTFS',
                                  'etc', 'init', 'setup-stopped.conf')
        mojo.utils.mkdirs(os.path.dirname(setup_file))
        open(setup_file, 'a').close()

        run_with_args(args)

        expected_dirs = ['build', 'local', 'log', 'spec']
        workspace_dir = os.path.join(mojo_root, 'my-project', 'precise',
                                     'my-workspace')
        for expected_dir in expected_dirs:
            expected_path = os.path.join(workspace_dir, expected_dir)
            self.assertTrue(os.path.exists(expected_path))

    @mock.patch('mojo.LXCContainer.is_setup', return_value=True)
    @mock.patch('mojo.LXCContainer.exists', return_value=True)
    @mock.patch('mojo.Project.list', return_value='my-project')
    def test_workspace_new_env_vars(self, project_list, exists, is_setup):

        mojo_root = self.patch_mojo_root()
        spec_dir = self.create_spec_dir()
        parser = create_arg_parser()
        args = parser.parse_args([
            "workspace-new",
            spec_dir,
            "my-workspace",
        ])

        # Mock the LXC directory
        lxc_project_series = os.path.join(mojo_root, 'your-project', 'precise',
                                          'ROOTFS', mojo_root[1:],
                                          'your-project', 'precise')
        mojo.utils.mkdirs(lxc_project_series)
        setup_file = os.path.join(mojo_root, 'your-project', 'precise',
                                  'ROOTFS', 'etc', 'init', 'setup-stopped.conf')
        mojo.utils.mkdirs(os.path.dirname(setup_file))
        open(setup_file, 'a').close()

        with use_env(MOJO_PROJECT='your-project', MOJO_SERIES='precise'):
                # Mock Project.list()
                with mock.patch.object(mojo.Project, 'list',
                                       return_value='your-project'):
                    run_with_args(args)

        expected_dirs = ['build', 'local', 'log', 'spec']
        workspace_dir = os.path.join(mojo_root, 'your-project', 'precise',
                                     'my-workspace')
        for expected_dir in expected_dirs:
            expected_path = os.path.join(workspace_dir, expected_dir)
            self.assertTrue(os.path.exists(expected_path))
